"""Module to test pipeline telemetry helper methods.
"""
import test_data as td
from errors import ReturnValueWithErrorStatus, ReturnValueWithStatus

import pipeline_telemetry.helper as HP
from pipeline_telemetry import Telemetry, add_errors_from_return_value, \
    increase_base_count, increase_fail_count, process_return_value, \
    process_telemetry_counters_in_return_value
from pipeline_telemetry.settings.settings import BASE_COUNT_KEY, FAIL_COUNT_KEY


class HelperTest:
    pass


def test_add_result_value_with_status_without_errors():
    """
    Test adding the ResultValueWithStatus instance without an error in it, to
    the telemetry adds no errors to the telemetry object.
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    return_value_without_error = ReturnValueWithStatus(result=[])
    add_errors_from_return_value(
        object_with_telemetry=test_obj,
        sub_process="RETRIEVE_RAW_DATA",
        return_value=return_value_without_error)
    assert not test_obj._telemetry.get("RETRIEVE_RAW_DATA").get('errors')


def test_add_result_value_with_status_with_one_error():
    """
    Test adding the ResultValueWithStatus instance with an error in it to the
    telemetry adds the actual error to the telemetry object.
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    return_value_with_error = ReturnValueWithErrorStatus(
        error=td.TEST_ERROR_CODE)
    add_errors_from_return_value(
        object_with_telemetry=test_obj,
        sub_process="RETRIEVE_RAW_DATA",
        return_value=return_value_with_error)
    assert td.TEST_ERROR_CODE.code in \
        str(test_obj._telemetry.get("RETRIEVE_RAW_DATA").get('errors'))
    assert test_obj._telemetry.get("RETRIEVE_RAW_DATA").get('errors').get(
        td.TEST_ERROR_CODE.code) == 1


def test_add_result_value_with_status_with_errors():
    """
    Test adding the ResultValueWithStatus instance with 2 errors in it to the
    telemetry adds the actual errors to the telemetry object.
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    return_value_with_error = ReturnValueWithErrorStatus(
        error=td.TEST_ERROR_CODE)
    return_value_with_error.add_error(td.TEST_ERROR_CODE)
    add_errors_from_return_value(
        object_with_telemetry=test_obj,
        sub_process="RETRIEVE_RAW_DATA",
        return_value=return_value_with_error)
    assert test_obj._telemetry.get("RETRIEVE_RAW_DATA").get('errors').get(
        td.TEST_ERROR_CODE.code) == 2


def test_process_telemetry_counters_in_return_value():
    """
    Test that when a ResultValueWithStatus instance with 2 telemetry counters
    in the result is processed by process_telemetry_counters_in_return_value
    the telemetry counters are processed.
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    process_telemetry_counters_in_return_value(
        object_with_telemetry=test_obj,
        return_value=td.TEST_RETURN_VALUE)
    assert test_obj._telemetry.get("RETRIEVE_RAW_DATA").get(
        'errors') is not None
    assert test_obj._telemetry.get("RETRIEVE_RAW_DATA").get('test_counter') == 1


def test_increase_base_count():
    """
    Test increase base count without increment increase base counter with 1 for
    a given sub process.
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    increase_base_count(
        object_with_telemetry=test_obj,
        sub_process="RETRIEVE_RAW_DATA")
    assert test_obj._telemetry.get("RETRIEVE_RAW_DATA").get(BASE_COUNT_KEY) == 1


def test_increase_base_count_with_increment():
    """
    Test increase base count with increment 2 increases base counter with 2 for
    a given sub process.
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    increase_base_count(
        object_with_telemetry=test_obj,
        sub_process="RETRIEVE_RAW_DATA",
        increment=2)
    assert test_obj._telemetry.get("RETRIEVE_RAW_DATA").get(BASE_COUNT_KEY) == 2


def test_increase_fail_count():
    """
    Test increase fail count without increment increases teh fail counter with
    1 for a given sub process.
    """
    test_obj = HelperTest()
    test_obj._telemetry = Telemetry(**td.DEFAULT_TELEMETRY_PARAMS)
    test_obj._telemetry._initialize_sub_process("RETRIEVE_RAW_DATA")
    increase_fail_count(
        object_with_telemetry=test_obj,
        sub_process="RETRIEVE_RAW_DATA")
    assert test_obj._telemetry.get("RETRIEVE_RAW_DATA").get(FAIL_COUNT_KEY) == 1


def test_process_return_value(mocker):
    """
    Test that process_return_value returns the list from the return_value
    result field but with all TelemetryCounter instances removed.
    """
    mocker.patch(
        ("pipeline_telemetry.helper."
         "process_telemetry_counters_in_return_value"),
        return_value=None)
    mocker.patch(
        ("pipeline_telemetry.helper."
         "add_errors_from_return_value"),
        return_value=None)
    assert process_return_value(
        object_with_telemetry=None,
        sub_process="na for this test",
        return_value=td.TEST_RETURN_VALUE) == [1, 2]


def test_process_return_value_processes_the_errors(mocker):
    """
    Test process_return_value calls the add_errors_from_return_value method.
    """
    mocker.patch(
        ("pipeline_telemetry.helper."
         "process_telemetry_counters_in_return_value"),
        return_value=None)
    mocker.patch(
        ("pipeline_telemetry.helper."
         "add_errors_from_return_value"),
        return_value=None)
    process_telemetry_spy = mocker.spy(
        HP, "process_telemetry_counters_in_return_value")

    process_return_value(
        object_with_telemetry=None,
        sub_process="na for this test",
        return_value=td.TEST_RETURN_VALUE)
    assert process_telemetry_spy.called


def test_process_return_value_processes_the_telemetry_counters(mocker):
    """
    Test process_return_value calls the
    process_telemetry_counters_in_return_value method.
    """
    mocker.patch(
        ("pipeline_telemetry.helper."
         "process_telemetry_counters_in_return_value"),
        return_value=None)
    mocker.patch(
        ("pipeline_telemetry.helper."
         "add_errors_from_return_value"),
        return_value=None)
    process_telemetry_spy = mocker.spy(
        HP, "process_telemetry_counters_in_return_value")

    process_return_value(
        object_with_telemetry=None,
        sub_process="na for this test",
        return_value=td.TEST_RETURN_VALUE)
    assert process_telemetry_spy.called



"""
Module to provide Telemetry class as public class to be accessed

classes
    - Telemetry

"""
from collections import defaultdict
from datetime import datetime
from typing import List

from errors import ErrorCode

from .helper import _raise_exception_if_telemetry_closed
from .settings import exceptions
from .settings import settings as st
from .settings.data_class import ProcessType, TelemetryCounter
from .settings.process_type import ProcessTypes
from .storage.generic import AbstractTelemetryStorage
from .storage.memory import TelemetryInMemoryStorage
from .validators.dict_validator import DictValidator


class Telemetry:
    """Class to manage the telemetry data object of a data pipeline process.

    This class can be used to measure and store indicators of a data process.
    When the dataprocess is finished this Telemetry object can be persisted in
    a database provide via a storage class.

    args:
        - source_name (str):
            Process name for which telemetry is made for example: `GET-WEATHER`
            Process names can be choosen freely and should ne unique for a
            single process as they are used to collect all telemetry data for a
            specific process.
        - process_type (ProcessType):
            Process Type for which the telemetry is created. For example
            `CREATE_DATA_FROM_API`. Process type are predefined objects and
            made available via class ProcessTypes. Each process type defines
            a number of sub processes like for example `RETRIEVE_RAW_DATA`. For
            these subprocess specific telemetry data can be stored in the
            telemetry instance
        - telemetry_rules (dict)
            rules how to evaluate the data object for specific sub process
            see DictValidator class for more details on how to define rules
        - storage_class (AbstractTelemetryStorage)
            class to define storage method for Telemetry object

    public class methods:
        - add_process_types: add custom process types

    properties:
        - category
        - sub_category
        - source_name
        - process_type
        - sub_process_types

    public methods:
        - get (as normal get method for dict)
        - save_and_close
        - add
        - increase_sub_process_base_count
        - increase_sub_process_fail_count
        - increase_sub_process_custom_count
        - increase_custom_count
    """

    _available_process_types = ProcessTypes
    _available_telemetry_types = st.TELEMETRY_TYPES

    def __init__(
        self,
        category: str,
        sub_category: str,
        source_name: str,
        process_type: ProcessType,
        telemetry_type: str = st.DEFAULT_TELEMETRY_TYPE,
        telemetry_rules: dict = dict(),
        storage_class: AbstractTelemetryStorage = None,
    ):
        self._set_process_type(process_type)
        self._telemetry = defaultdict(int)
        self._validate_telemetry_type(telemetry_type)
        self._telemetry.update({
            st.TELEMETRY_TYPE_KEY: telemetry_type,
            st.CATEGORY_KEY: category,
            st.SUB_CATEGORY_KEY: sub_category,
            st.SOURCE_NAME_KEY: source_name,
            st.PROCESS_TYPE_KEY: self._process_type.name,
            st.START_TIME: datetime.now(),
            st.IO_TIME_KEY: 0,
            st.TRAFFIC_LIGHT_KEY: st.DEFAULT_TRAFIC_LIGHT_COLOR
        })
        self._telemetry_rules = telemetry_rules
        self._storage_class = self._get_storage_class(storage_class)

    @classmethod
    def add_process_type(cls, process_type_key: str, process_type: ProcessType) -> None:
        """
        Add a custom process type to the available process types to the
        already registered process types.

        Args:
            process_type (ProcessType): Process type that needs to be added
            process_type_key (str): key that should be used when calling this
                                    ProcessType
        """
        if not isinstance(process_type, ProcessType):
            raise exceptions.ProcessTypeMustBeOfClassProcessType()
        cls._available_process_types.register_process_type(
            process_type_key, process_type
        )

    def _validate_telemetry_type(self, telemetry_type: str) -> None:
        """Check validaty of provided telemetry type.

        Raises exception if not valid. Returns non if telemetry type is valid.

        Args:
            telemetry_type (str): [description]

        Raises:
            exceptions.InvalidTelemetryType: When telemetry type is not valid.
        """

        if telemetry_type not in self._available_telemetry_types:
            raise exceptions.InvalidTelemetryType(
                self._available_telemetry_types)

    @staticmethod
    def _get_storage_class(storage_class: AbstractTelemetryStorage) -> None:
        """
        validates and sets the storage class. If no storage class provided
        it is defaulted to TelemetryInMemoryStorage

        Args:
            storage_class (AbstractTelemetryStorage):
                Storage class, child class from AbstractTelemetryStorage

        Raises:
            StorageClassOfIncorrectType
        """
        # set default in memory storage if no storage class provided
        if not storage_class:
            storage_class = TelemetryInMemoryStorage

        if not issubclass(storage_class, AbstractTelemetryStorage):
            raise exceptions.StorageClassOfIncorrectType(
                storage_class.__class__.__name__
            )

        return storage_class()

    @property
    def storage_class(self):
        return self._storage_class

    @storage_class.setter
    def storage_class(self, storage_class):
        self._storage_class = self._get_storage_class(storage_class)

    @property
    def source_name(self) -> str:
        """Source_name property."""
        return self._telemetry.get(st.SOURCE_NAME_KEY)

    @property
    def traffic_light(self) -> str:
        """Traffic_light property."""
        return self._telemetry.get(st.TRAFFIC_LIGHT_KEY)

    @property
    def telemetry(self) -> dict:
        """Telemetry property."""
        return self._telemetry

    @property
    def sub_process_types(self) -> list:
        """Returns a of subprocess types allowed for the Telemetry instance."""
        return self._process_type.sub_processes

    def set_orange_traffic_light(self) -> None:
        """Sets traffic light attribute to orange."""
        self._telemetry[st.TRAFFIC_LIGHT_KEY] = st.TRAFIC_LIGHT_COLOR_ORANGE

    def set_red_traffic_light(self) -> None:
        """Sets traffic light attribute to red."""
        self._telemetry[st.TRAFFIC_LIGHT_KEY] = st.TRAFIC_LIGHT_COLOR_RED

    def get(self, field_name: str) -> dict:
        """Retrieve a field from the telemetry object.

        Implements the get method on the Telemetry object as if it is a dict

        Returns:
            dict: [description]
        """
        return self._telemetry.get(field_name)

    def save_and_close(self) -> dict:
        """
        Closes and stores the telemetry instance and returns the
        telemetry value.

        Returns:
            dict: telemetry result
        """
        if self._telemetry.get(st.RUN_TIME):
            raise exceptions.TelemetryObjectAlreadyClosed()
        self._set_runtime()
        self._storage_class.store_telemetry(self._telemetry)

        return self._telemetry

    def _set_runtime(self) -> None:
        """Method to calculate and set the runtime seconds (in str)."""
        run_time = datetime.now() - self._telemetry.get(st.START_TIME)
        self._telemetry[st.RUN_TIME] = str(run_time.total_seconds())

    @_raise_exception_if_telemetry_closed
    def add(self, sub_process: str, data: dict, errors: List[ErrorCode]) -> None:
        """
        Add data validation errors and/or errors from data process to a
        telemetry sub process.
        Data validation errors are retrieved from data validation defined in
        telemetry rules.

        Method does not update BASE_COUNT.

        :param sub_process: applicable subprocess for the errors
        :type sub_process: str
        :param data: data generated by subproces for which telemetry
                     needs to be generated
        :param type: dict
        :param errors: errors from data process to be added
        :type errors: list of Errorcodes
        :returns: None
        """
        if self._sub_process_not_yet_initialized(sub_process):
            self._initialize_sub_process(sub_process)
        validation_errors = self._validate_data(sub_process, data)
        # add data validation errors
        self._add_errors(sub_process, validation_errors)

        # add data process errors
        if errors:
            self._add_errors(sub_process, errors)

    def _validate_data(self, sub_process: str, data: dict) -> List[ErrorCode]:
        """Validates the data provided by a subprocess.

        Data is validated based upon telemetry rules for the subprocces

        :param sub_process: applicable subprocess
        :typesub_process: str
        :param data: data provided by subprocess
        :type data: dict
        :returns: list of Errorcodes
        """
        telemetry_rules = self._telemetry_rules.get(sub_process, {})
        return DictValidator().validate(data, telemetry_rules)

    def _add_errors(self, sub_process: str, errors: list) -> None:
        """Adds the error to a telemetry sub process.

        :param sub_process: applicable subprocess for the errors
        :typesub_process: str
        :param errors: errors to be processed
        :type errors: list of Errorcodes
        :returns: None
        """
        for error in errors:
            self.increase_sub_process_error_count(error=error, sub_process=sub_process)

    @_raise_exception_if_telemetry_closed
    def increase_sub_process_base_count(
        self, sub_process: str, increment: int = 1
    ) -> None:
        """
        Increases the base count for a subprocess.

        Args:
            sub_process (str): name of subprocess
        """
        if self._sub_process_not_yet_initialized(sub_process):
            self._initialize_sub_process(sub_process)

        self._telemetry[sub_process][st.BASE_COUNT_KEY] += increment

    @_raise_exception_if_telemetry_closed
    def increase_sub_process_fail_count(
            self, sub_process: str, increment: int = 1) -> None:
        """Increases the fail count for a subprocess

        Args:
            sub_process (str): name of subprocess
        Raises:
            BaseCountForSubProcessNotAdded: if subprocess has not yet been
                                            created
        """
        if self._sub_process_not_yet_initialized(sub_process):
            raise exceptions.BaseCountForSubProcessNotAdded(sub_process)

        self._telemetry[sub_process][st.FAIL_COUNT_KEY] += increment

    @_raise_exception_if_telemetry_closed
    def increase_sub_process_custom_count(
        self, custom_counter: str, sub_process: str, increment: int = 1
    ) -> None:
        """
        Increases a custom counter for a subprocess.

        Args:
            sub_process (str): name of subprocess
            custom_counter (str): name of custom counter
            increment (int): increment for the counter (default = 1)

        Raises:
            BaseCountForSubProcessNotAdded: if subprocess has not yet been
                                            created
        """
        if self._sub_process_not_yet_initialized(sub_process):
            raise exceptions.BaseCountForSubProcessNotAdded(sub_process)

        self._telemetry[sub_process][custom_counter] += increment

    @_raise_exception_if_telemetry_closed
    def increase_sub_process_error_count(
        self, sub_process: str, error: ErrorCode, increment: int = 1
    ) -> None:
        """
        Increases a error counter for a subprocess.

        Args:
            sub_process (str): name of subprocess
            error_code (str): error code from ErrorCode object
            increment (int): increment for the counter

        Raises:
            BaseCountForSubProcessNotAdded: if subprocess has not yet been
                                            created
        """
        if self._sub_process_not_yet_initialized(sub_process):
            raise exceptions.BaseCountForSubProcessNotAdded(sub_process)

        self._telemetry[sub_process][st.ERRORS_KEY][error.code] += increment

    @_raise_exception_if_telemetry_closed
    def increase_io_time(self, io_time: float) -> None:
        """
        Increases the io_time of the telemetry object.

        Args:
            io_time (float): io_time that needs to be added to the total io_time
        """
        self._telemetry[st.IO_TIME_KEY] += io_time

    @_raise_exception_if_telemetry_closed
    def increase_custom_count(self, custom_counter: str, increment: int = 1) -> None:
        """
        Increases a custom counter in the top level of the telemetry object.
        When then custom counter does not yet exist is created.

        Args:
            custom_counter (str): name of custom counter
            increment (int): increment for the counter (Default = 1)
        """
        self._telemetry[custom_counter] += increment

    def add_telemetry_counter(
        self, telemetry_counter: TelemetryCounter, increment: int = None
    ) -> None:
        """
        Method to process a TelemetryCounter object with predefined counters.

        Args:
            telemetry_counter (TelemetryCounter): [description]
            increment (int, None):
                when not None overules the increment setting from
                TelemetryCounter (which defaults to 1)
        """
        telemetry_counter.validate_sub_process()
        sub_process = telemetry_counter.sub_process
        counter_name = telemetry_counter.counter_name
        increment = increment or telemetry_counter.increment

        if self._sub_process_not_yet_initialized(sub_process):
            self._initialize_sub_process(sub_process)

        if telemetry_counter.error:
            self.increase_sub_process_error_count(
                sub_process=sub_process,
                error=telemetry_counter.error,
                increment=increment,
            )

        else:
            self.increase_sub_process_custom_count(
                sub_process=sub_process,
                custom_counter=counter_name,
                increment=increment,
            )

    def _sub_process_not_yet_initialized(self, sub_process: str) -> bool:
        """Returns True if provided sub_process is not yet initialized

        Args:
            sub_process (str): sub_process

        Returns:
            bool: True if sub_process is inialized, False otherwis
        """
        return not self._sub_process_is_initialized(sub_process)

    def _sub_process_is_initialized(self, sub_process: str) -> bool:
        """Returns True if provided sub_process is initialized

        Args:
            sub_process (str): sub_process

        Returns:
            bool: True if sub_process is inialized, False otherwis
        """
        return self._telemetry.get(sub_process)

    def _initialize_sub_process(self, sub_process: str) -> None:
        """sets the initial count object for a sub_process

        Args:
            sub_process (str): sub process name
        """
        if sub_process not in self._process_type.sub_processes:
            raise exceptions.InvalidSubProcess(sub_process, self._process_type)

        self._telemetry[sub_process] = defaultdict(int)
        self._telemetry[sub_process].update({
            st.BASE_COUNT_KEY: 0,
            st.FAIL_COUNT_KEY: 0,
            st.ERRORS_KEY: defaultdict(int)
        })

    def _set_process_type(self, process_type: ProcessType) -> None:
        """Sets the process type for the telemetrty instance."""
        if not isinstance(process_type, ProcessType):
            raise exceptions.ProcessTypeMustBeOfClassProcessType
        if not self._available_process_types.is_registered(process_type):
            raise exceptions.ProcessTypeNotRegistered(process_type)
        self._process_type = process_type



"""Telemety Decorators to be used whith class methods that need telemetry

decorators:
    - add_telemetry
    - add_mongo_telemetry
    - add_single_usage_telemetry
    - add_mongo_single_usage_telemetry
"""
from functools import wraps

from .main import Telemetry
from .settings import exceptions
from .storage import AbstractTelemetryStorage, TelemetryMongoStorage


def add_telemetry(telemetry_params: dict) -> object:
    """
    Decorator method to add a telemetry to the class from which the
    decorator was called. When using this decorator the storage
    class can be set as wanted but defaults to TelemetryInMemoryStorage.

    Args:
        telemetry_params (dict:
            - source_name (str): free format process name
            - process_type (ProcessType): process type definition
            - telemetry_rules (dict): telemetry rules definition dict
            - storage_class (AbstractTelemetryStorage, optional):
                Storage class to be used to store telemetry instances. Defaults
                to TelemetryInMemoryStorage which stores only in memory.
    """

    def wrapper(method):
        @wraps(method)
        def wrapped_method(self, *args, **kwargs):
            """
            Wrapper for method where result log should be added
            """
            if (not hasattr(self, "_telemetry")) or (not self._telemetry):
                self._telemetry = Telemetry(**telemetry_params)
                result = method(self, *args, **kwargs)
                self._telemetry.save_and_close()
                self._telemetry = None
            else:
                result = method(self, *args, **kwargs)

            return result

        return wrapped_method

    return wrapper


def add_mongo_telemetry(telemetry_params: dict) -> object:
    """
    Decorator method to add a telemetry to the class from which the
    decorator was called. When using this decorator the mongo storage
    class will used.

    Args:
        telemetry_params (dict:
            - source_name (str): free format process name
            - process_type (ProcessType): process type definition
            - telemetry_rules (dict): telemetry rules definition dict
    """

    def wrapper(method):
        @wraps(method)
        def wrapped_method(self, *args, **kwargs):
            """
            Wrapper for method where result log should be added
            """
            if (not hasattr(self, "_telemetry")) or (not self._telemetry):
                self._telemetry = Telemetry(**telemetry_params)
                self._telemetry.storage_class = TelemetryMongoStorage
                result = method(self, *args, **kwargs)
                self._telemetry.save_and_close()
                self._telemetry = None
            else:
                result = method(self, *args, **kwargs)
            return result

        return wrapped_method

    return wrapper


def add_mongo_single_usage_telemetry(sub_process: str = None) -> object:
    """
    Decorator method to add a telemetry to the class from which the
    decorator was called. The telemetry object will be reset each the method
    initiating the telemetry object will be called. This method should be used
    for stand alone telemetry measurements like for example external API calls.
    It can be used to easily store a single event telemetry object.

    This method will persist the telemetry object in MongoDB

    Args:
        - sub_process (str, Optional):
            one of the sub_processes defined with the process_type. A
            counter with value 1 will be created for this sub_process.
            This field is optional, if not provided no counter will be
            added to the telemetry object
    """
    return add_single_usage_telemetry(
        sub_process=sub_process,
        storage_class=TelemetryMongoStorage
    )


def add_single_usage_telemetry(
        sub_process: str = None,
        storage_class: AbstractTelemetryStorage = None) -> object:
    """
    Decorator method to add a telemetry to the class from which the
    decorator was called. The telemetry object will be reset each the method
    initiating the telemetry object will be called. This method should be used
    for stand alone telemetry measurements like for example external API calls.
    It can be used to easily store a single event telemetry object.

    Args:
        - sub_process (str):
            one of the sub_processes defined with the process_type. A
            counter with value 1 will be created for this sub_process.
            This field is optional, if not provided no counter will be
            added to the telemetry object
        - storage_class (AbstractTelemetryStorage):
            Storage class to be used for persisting telemetry objects
    """

    def wrapper(method):
        @wraps(method)
        def wrapped_method(self, *args, **kwargs):
            """
            Wrapper for method where result log should be added
            """
            telemetry_params = getattr(self, 'TELEMETRY_PARAMS', False)
            if not telemetry_params:
                raise exceptions.ClassTelemetryParamsNotDefined(self)

            self._telemetry = Telemetry(**telemetry_params)

            if storage_class:
                self._telemetry.storage_class = storage_class

            # only if sub_process was defined set the base count for that
            # subprocess
            if sub_process:
                self._telemetry.increase_sub_process_base_count(
                    sub_process=sub_process)
            result = method(self, *args, **kwargs)
            self._telemetry.save_and_close()
            self._telemetry = None

            return result

        return wrapped_method

    return wrapper



"""[summary]
"""
import json
import sqlite3

from ..settings import settings as st
from .generic import AbstractTelemetryStorage


class TelemetryInMemoryStorage(AbstractTelemetryStorage):
    """
    Class to provice telemetry in memory storage for use when unit testing
    """

    db_in_memory = None
    db_cursor = None

    def __init__(self):
        if not self.db_in_memory:
            self.initialize_db()

    @classmethod
    def initialize_db(cls):
        """
        class method to initialize the in memory db
        """
        if not cls.db_in_memory:
            cls.db_in_memory = sqlite3.connect(":memory:")
            cls.db_cursor = cls.db_in_memory.cursor()
            cls._define_db_table(cls.db_cursor)

    @classmethod
    def close_db(cls):
        """close cursor and connections"""
        cls.db_cursor.close()
        cls.db_in_memory.close()
        cls.db_cursor = None
        cls.db_in_memory = None

    @staticmethod
    def _define_db_table(cursor):
        """define telemetry table"""
        cursor.executescript(
            """
            DROP TABLE IF EXISTS telemetry;
            CREATE TABLE telemetry (telemetry_type varchar(100),
            category varchar(60), sub_category varchar(60),
            source_name varchar(40), process_type varchar(40),
            start_date_time timestamp, run_time varchar(20),
            telemetry_data json, traffic_light varchar(10),
            io_time_in_seconds real)"""
        )

    def store_telemetry(self, telemetry: dict) -> None:
        """public method to persist telemetry object"""
        telemetry_copy = telemetry.copy()
        telemetry_type = telemetry_copy.pop(st.TELEMETRY_TYPE_KEY, None)
        category = telemetry_copy.pop(st.CATEGORY_KEY, None)
        sub_category = telemetry_copy.pop(st.SUB_CATEGORY_KEY, None)
        source_name = telemetry_copy.pop(st.SOURCE_NAME_KEY, None)
        process_type = telemetry_copy.pop(st.PROCESS_TYPE_KEY, None)
        start_date_time = telemetry_copy.pop(st.START_TIME, None)
        run_time_in_seconds = telemetry_copy.pop(st.RUN_TIME, None)
        traffic_light = telemetry_copy.pop(st.TRAFFIC_LIGHT_KEY, None)
        io_time_in_seconds = telemetry_copy.pop(st.IO_TIME_KEY, None)
        json_object = json.dumps(telemetry_copy)

        self.db_cursor.execute(
            "insert into telemetry values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
            [
                telemetry_type,
                category,
                sub_category,
                source_name,
                process_type,
                start_date_time,
                run_time_in_seconds,
                json_object,
                traffic_light,
                io_time_in_seconds
            ],
        )



"""
Module to define TelemetryMixin class for adding methods to
a class that allow easy Telemetry updates
"""
from errors import ReturnValueWithStatus

from .helper import add_errors_from_return_value


class TelemetryMixin():

    def process_errors_from_return_values(
            self,  sub_process: str,
            return_value: ReturnValueWithStatus) -> None:
        """
        Adds errors from return_value to _telemetry object for a specific
        sub_process.

        Args:
            sub_process (str):
                One of the sub_processed defined in process type for this
                telemetry object.
            return_value (ReturnValueWithStatus):
                Return value object containing the errors
        """
        add_errors_from_return_value(
            object_with_telemetry=self,
            sub_process=sub_process,
            return_value=return_value)



"""
Module to test telemetry mixin class for pipeline telemetry module.
"""
import pipeline_telemetry.mixin as HP
from pipeline_telemetry.mixin import TelemetryMixin


def test_telemetry_mixin_exists():
    """check that TelemetryMixin class exists"""
    assert TelemetryMixin


def test_process_errors_from_return_values(mocker):
    """
    Test method process_errors_from_return_value calls helper method
    add_errors_from_return_value
    """
    mocker.patch(
        ("pipeline_telemetry.mixin."
         "add_errors_from_return_value"),
        return_value=None)
    add_errors_spy = mocker.spy(
        HP, "add_errors_from_return_value")

    TelemetryMixin().process_errors_from_return_values(
        sub_process='TEST',
        return_value='test')
    assert add_errors_spy.called



"""[summary]
"""
from enum import Enum

from pipeline_telemetry.settings.data_class import ProcessType

DEFAULT_TELEMETRY_TYPE = 'SINGLE TELEMETRY'
TELEMETRY_TYPES = [
    'PARTIAL TELEMETRY', DEFAULT_TELEMETRY_TYPE, 'DAILY AGGREGATION',
    'WEEKLY AGGREGATION', 'MONTHLY AGGREGATION', 'QUARTERLY AGGREGATION']

telemetry_types = \
    {type.upper().replace(' ', '_'): type.upper() for type in TELEMETRY_TYPES}

TRAFIC_LIGHT_COLOR_GREEN = 'GREEN'
TRAFIC_LIGHT_COLOR_ORANGE = 'ORANGE'
TRAFIC_LIGHT_COLOR_RED = 'RED'
DEFAULT_TRAFIC_LIGHT_COLOR = TRAFIC_LIGHT_COLOR_GREEN


BASE_COUNT_KEY = "base_counter"
ERRORS_KEY = "errors"
FAIL_COUNT_KEY = "fail_counter"
SOURCE_NAME_KEY = "source_name"
CATEGORY_KEY = "category"
SUB_CATEGORY_KEY = "sub_category"
PROCESS_TYPE_KEY = "process_type"
START_TIME = "start_date_time"
RUN_TIME = "run_time_in_seconds"
TRAFFIC_LIGHT_KEY = "traffic_light"
TELEMETRY_TYPE_KEY = "telemetry_type"
IO_TIME_KEY = "io_time_in_seconds"
TELEMETRY_FIELD_KEY = "telemetry"

DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES = [
    "RETRIEVE_RAW_DATA",
    "DATA_CONVERSION",
    "DATA_STORAGE",
]

DEFAULT_UPLOAD_DATA_SUB_PROCESS_TYPES = [
    "DATA_SELECTION",
    "DATA_CONVERSION",
    "DATA_UPLOAD",
]


class BaseEnumerator(Enum):
    """
    Class to define base enumerator with a keys method.
    Only to be used for defining new enumerators.
    """

    @classmethod
    def keys(cls):
        return list(cls.__members__.keys())


class DefaultProcessTypes(BaseEnumerator):
    """
    Class to define the defailt process types with their subtypes
    """

    CREATE_DATA_FROM_URL = ProcessType(
        process_type="create_data_from_url",
        subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES,
    )

    CREATE_DATA_FROM_API = ProcessType(
        process_type="create_data_from_api",
        subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES,
    )

    CREATE_DATA_FROM_FILE = ProcessType(
        process_type="create_data_from_file",
        subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES,
    )

    UPLOAD_DATA = ProcessType(
        process_type="upload_data", subtypes=DEFAULT_UPLOAD_DATA_SUB_PROCESS_TYPES
    )



"""Module to test settings module.
"""
from pipeline_telemetry.settings import settings


def test_telemetry_type_settings():
    """
    Test that telemetry type settings is a dict where keys and values are
    in uppercase and where underscores in the keys are replaced by spaces
    in the values.
    """
    for k, v in settings.telemetry_types.items():
        assert k == k.upper()
        assert v == v.upper()
        assert v == k.replace('_', ' ')



"""Module to provide a storage class for using MongoDB.

When using MongoDB storage class you can select the mongoDB instance to be used
via environment Variables

MONGO_DB_NAME (defaults to `telemetry`)
MONGO_DB_PASSWORD
MONGO_DB_USERNAME
MONGO_DB_HOST
MONGO_DB_PORT

If no host and port are defined the connection will dedault to a localhost
mongoDB instance.
"""
from mongoengine import DateTimeField, DictField, Document, FloatField, \
    StringField, connect

from ..settings import settings as st
from .generic import AbstractTelemetryStorage
from .mongo_connection import MONGO_ACCESS_PARAMS

connect(alias="telemetry", **MONGO_ACCESS_PARAMS)


class TelemetryMongoModel(Document):
    """
    Class to provice telemetry Mongo Model for persistance in MongoDB
    """

    category = StringField()
    sub_category = StringField()
    source_name = StringField()
    process_type = StringField()
    start_date_time = DateTimeField()
    run_time_in_seconds = StringField()
    traffic_light = StringField()
    telemetry_type = StringField()
    io_time_in_seconds = FloatField(default=0)
    telemetry = DictField(default=None)

    meta = {
        "db_alias": "telemetry",
        "indexes": [
            "category",
            "sub_category",
            "source_name",
            ("category", "sub_category", "source_name", "process_type"),
            "process_type",
            "traffic_light",
            "start_date_time",
        ],
    }


class TelemetryMongoStorage(AbstractTelemetryStorage):
    """
    Class to provice telemetry in mongo storage class.
    This class can be used as storage_class argument when creating
    an instance of Telemetry.
    """

    def store_telemetry(self, telemetry: dict) -> None:
        """public method to persist telemetry object"""
        telemetry_mongo_kwargs = self._telemetry_model_kwargs(telemetry)
        TelemetryMongoModel(**telemetry_mongo_kwargs).save()

    @staticmethod
    def _telemetry_model_kwargs(telemetry: dict) -> dict:
        """
        Returns a dicts with kwargs that can be used to create a new
        TelemetryMongoStorage instance.
        """
        telemetry_copy = telemetry.copy()
        telemetry_type = telemetry_copy.pop(st.TELEMETRY_TYPE_KEY, None)
        category = telemetry_copy.pop(st.CATEGORY_KEY, None)
        sub_category = telemetry_copy.pop(st.SUB_CATEGORY_KEY, None)
        source_name = telemetry_copy.pop(st.SOURCE_NAME_KEY, None)
        process_type = telemetry_copy.pop(st.PROCESS_TYPE_KEY, None)
        start_date_time = telemetry_copy.pop(st.START_TIME, None)
        run_time_in_seconds = telemetry_copy.pop(st.RUN_TIME, None)
        traffic_light = telemetry_copy.pop(st.TRAFFIC_LIGHT_KEY, None)
        io_time_in_seconds = telemetry_copy.pop(st.IO_TIME_KEY, None)

        return {
            st.TELEMETRY_TYPE_KEY: telemetry_type,
            st.CATEGORY_KEY: category,
            st.SUB_CATEGORY_KEY: sub_category,
            st.SOURCE_NAME_KEY: source_name,
            st.PROCESS_TYPE_KEY: process_type,
            st.START_TIME: start_date_time,
            st.RUN_TIME: run_time_in_seconds,
            st.TRAFFIC_LIGHT_KEY: traffic_light,
            st.TELEMETRY_FIELD_KEY: telemetry_copy,
            st.IO_TIME_KEY: io_time_in_seconds,
        }



"""Helper module for pipeline_telemetry
"""
from typing import Any, List, Union

from errors import ReturnValueWithStatus

from .settings import exceptions
from .settings import settings as st
from .settings.data_class import TelemetryCounter


def is_telemetry_counter(counter: Union[TelemetryCounter, Any]) -> bool:
    """Method to check in object is an instance of TelemetryCounter."""
    return issubclass(counter.__class__, TelemetryCounter)


def add_errors_from_return_value(
        object_with_telemetry: Any, sub_process: str,
        return_value: ReturnValueWithStatus) -> None:
    """
    Helper method to add the errors from a ReturnValueWithStatus instance to the
    telemetry instance of the object_with_telemetry.
    """
    object_with_telemetry._telemetry.add(
        sub_process=sub_process,
        data=[], errors=return_value.errors)


def process_telemetry_counters_in_return_value(
        object_with_telemetry: Any,
        return_value: ReturnValueWithStatus) -> None:
    """
    Helper method to add the errors from a ReturnValueWithStatus instance to the
    telemetry instance of the object_with_telemetry.
    """
    for item in return_value.result:
        if is_telemetry_counter(item):
            object_with_telemetry._telemetry.add_telemetry_counter(item)


def process_return_value(
        object_with_telemetry: Any, sub_process: str,
        return_value: ReturnValueWithStatus) -> List[Any]:
    """Processes a return value object.
    All errors and telemetry counters in return_value errors and result list
    will be processed and added to the telemetry object in
    object_with_telemetry. All the other values in the result list are then
    returned as a list.

    Args:
        object_with_telemetry (Any): _description_
        sub_process (str): _description_
        return_value (ReturnValueWithStatus): _description_

    Returns:
        List[Any]: _description_
    """
    add_errors_from_return_value(
        object_with_telemetry, sub_process, return_value)

    process_telemetry_counters_in_return_value(
        object_with_telemetry, return_value)

    return [
        item for item in return_value.result if not is_telemetry_counter(item)]


def increase_base_count(
    object_with_telemetry: Any, sub_process: str, increment: int = 1
) -> None:
    """
    Helper method to increase base count for a sub_process
    """
    object_with_telemetry._telemetry.increase_sub_process_base_count(
        sub_process=sub_process, increment=increment
    )


def increase_fail_count(
    object_with_telemetry: Any, sub_process: str, increment: int = 1
) -> None:
    """
    Helper method to increase base count for a sub_process
    """
    object_with_telemetry._telemetry.increase_sub_process_fail_count(
        sub_process=sub_process, increment=increment
    )


# decorator method to check status telemetry object
def _raise_exception_if_telemetry_closed(method):
    """
    Decorator method to check if telemetry object is closed.
    If so an exception is raised.

    Decorator method to be used for methods that are only allowed
    when telemetry object not yet closed.
    """

    def wrapper(self, *args, **kwargs):
        """
        Wrapper to check if run_time has been set
        If so the telemetry object is closed
        """
        if self.telemetry.get(st.RUN_TIME):
            raise exceptions.TelemetryObjectAlreadyClosed()

        return method(self, *args, **kwargs)

    return wrapper
